unit CTT;

interface

uses xmldom, XMLIntf, msxmldom, XMLDoc;

type
TTemporalOperator = (Concurrent, Choice, ConcerrentWithInfoExchange, Disabling, Enabling,
                     EnablingWithInfoExchange, SuspendResume, OrderIndependant, Null);
TTaskCategory = (Abstraction, App, Interaction, User);
TCTTNode = class
  private
    fName : String;
    fTemporalOperator : TTemporalOperator;
    fParent : TCTTNode;
    fChild : TCTTNode;
    fSiblingLeft : TCTTNode;
    fSiblingRight : TCTTNode;
    fOptional : Boolean;
    fIterative : Boolean;
    fCategory : TTaskCategory;
    fEnabled: Boolean;
    fPerformed: Boolean;
  public
    property Name : String
        read fName
        write fName;
    property TemporalOperator : TTemporalOperator
        read fTemporalOperator
        write fTemporalOperator;
    property Parent : TCTTNode
        read fParent
        write fParent;
    property Child : TCTTNode
        read fChild
        write fChild;
    property SiblingLeft : TCTTNode
        read fSiblingLeft
        write fSiblingLeft;
    property SiblingRight : TCTTNode
        read fSiblingRight
        write fSiblingRight;
    property Optional : Boolean
        read fOptional
        write fOptional;
    property Iterative : Boolean
        read fIterative
        write fIterative;
    property Category : TTaskCategory
        read fCategory
        write fCategory;
    property Enabled : Boolean
        read fEnabled
        write fEnabled;
    property Performed : Boolean
        read fPerformed
        write fPerformed;

    constructor Create; overload;
    function HasChild : Boolean;
    function HasSiblingRight : Boolean;
    function HasSiblingLeft : Boolean;
    function TaskCount : Integer;
    function FindChildByName(TaskName : String) : TCTTNode;
  
end;

function ImportCTTModel(Model : TXMLDocument) : TCTTNode;
function ImportCTTNode(CurrentNode : IXMLNode) : TCTTNode;
function ImportCTTSubTask(CurrentNode : IXMLNode; Parent : TCTTNode) : TCTTNode;
function StringToCategory(Category : String) : TTaskCategory;
function StringToTemporalOperator(TemporalOperator : String) : TTemporalOperator;


implementation

{Import Functions}

function ImportCTTModel(Model : TXMLDocument) : TCTTNode;
var
  xmlnodeTaskModel, xmlnodeCurrentTask, xmlnodeSubTask : IXMLNode;
begin
  xmlnodeTaskModel := Model.DocumentElement;
  xmlnodeCurrentTask := xmlnodeTaskModel.ChildNodes['Task'];
  Result := ImportCTTNode(xmlnodeCurrentTask);
  if xmlnodeCurrentTask.ChildNodes.IndexOf('SubTask') > -1 then
  begin
    xmlnodeSubTask := xmlnodeCurrentTask.ChildNodes['SubTask'];
    Result.Child := ImportCTTSubTask(xmlnodeSubTask, Result);
  end;
end;


function StringToCategory(Category: String): TTaskCategory;
begin
  if Category = 'application' then
    Result := App
  else if Category = 'user' then
    Result := User
  else if Category = 'interaction' then
    Result := Interaction
  else
    Result := Abstraction;
end;

function StringToTemporalOperator(
  TemporalOperator: String): TTemporalOperator;
begin
  if TemporalOperator = 'Choice' then
    Result := Choice
  else if TemporalOperator = 'OrderIndependence' then
    Result := OrderIndependant
  else if TemporalOperator = 'Synchronization' then
    Result := ConcerrentWithInfoExchange
  else if TemporalOperator = 'Interleaving' then
    Result := Concurrent
  else if TemporalOperator = 'Disabling' then
    Result := Disabling
  else if TemporalOperator = 'SuspendResume' then
    Result := SuspendResume
  else if TemporalOperator = 'SequentialEnabling' then
    Result := Enabling
  else if TemporalOperator = 'SequentialEnablingInfo' then
    Result := EnablingWithInfoExchange
  else
    Result := Null;
end;

function ImportCTTNode(CurrentNode: IXMLNode): TCTTNode;
var
  sCategory, sIterative, sOptional,
  sTemporalOperator : String;
begin
  Result := TCTTNode.Create;
  Result.Name := CurrentNode.Attributes['Identifier'];

  sCategory := CurrentNode.Attributes['Category'];
  Result.Category := StringToCategory(sCategory);

  sIterative := CurrentNode.Attributes['Iterative'];
  if sIterative = 'false' then
    Result.Iterative := false
  else
    Result.Iterative := true;

  sOptional := CurrentNode.Attributes['Optional'];
  if sOptional = 'false' then
    Result.Optional := false
  else
    Result.Optional := true;

  if CurrentNode.ChildNodes.IndexOf('TemporalOperator') > -1 then
  begin
    sTemporalOperator := CurrentNode.ChildNodes['TemporalOperator'].Attributes['name'];
    Result.TemporalOperator := StringToTemporalOperator(sTemporalOperator);
  end
  else
  begin
    Result.TemporalOperator := Null;
  end;
end;

function ImportCTTSubTask(CurrentNode: IXMLNode; Parent: TCTTNode): TCTTNode;
var
  i : Integer;
  cttLastNode, cttCurNode : TCTTNode;
  xmlnodeSubTask : IXMLNode;
begin

  Result := ImportCTTNode(CurrentNode.ChildNodes[1]);
  cttCurNode := Result;
  cttCurNode.Parent := Parent;
  if CurrentNode.ChildNodes[1].ChildNodes.IndexOf('SubTask') > -1 then
  begin
    xmlnodeSubTask := CurrentNode.ChildNodes[1].ChildNodes['SubTask'];
    cttCurNode.Child := ImportCTTSubTask(xmlnodeSubTask, cttCurNode);
  end;
  for i:=2 to CurrentNode.ChildNodes.Count-1 do
  begin
    cttLastNode := cttCurNode;
    cttCurNode := ImportCTTNode(CurrentNode.ChildNodes[i]);
    cttCurNode.Parent := Parent;
    cttLastNode.SiblingRight := cttCurNode;
    cttCurNode.SiblingLeft := cttLastNode;
    if CurrentNode.ChildNodes[i].ChildNodes.IndexOf('SubTask') > -1 then
    begin
      xmlnodeSubTask := CurrentNode.ChildNodes[i].ChildNodes['SubTask'];
      cttCurNode.Child := ImportCTTSubTask(xmlnodeSubTask, cttCurNode);
    end;
  end;
end;


{ TCTTNode }

constructor TCTTNode.Create;
begin
  inherited;
  fName := '';
  fTemporalOperator := Null;
  fParent := nil;
  fChild := nil;
  fSiblingLeft := nil;
  fSiblingRight := nil;
  fOptional := false;
  fIterative := false;
  fCategory := Abstraction;
  fEnabled := true;
  fPerformed := false;
end;

function TCTTNode.FindChildByName(TaskName: String): TCTTNode;
begin
  if fName = TaskName then
    Result := self
  else if (fChild = nil) AND (fSiblingRight = nil) then
    Result := nil
  else if (not (fSiblingRight = nil)) AND (fChild = nil) then
    Result := fSiblingRight.FindChildByName(TaskName)
  else if (not (fChild = nil)) AND (fSiblingRight = nil) then
    Result := fChild.FindChildByName(TaskName)
  else
  begin
    Result := fChild.FindChildByName(TaskName);
    if Result = nil then
      Result := fSiblingRight.FindChildByName(TaskName);
  end;
end;

function TCTTNode.HasChild: Boolean;
begin
  if fChild = nil then
    Result := false
  else
    Result := true;
end;

function TCTTNode.HasSiblingLeft: Boolean;
begin
  if fSiblingLeft = nil then
    Result := false
  else
    Result := true;
end;

function TCTTNode.HasSiblingRight: Boolean;
begin
  if fSiblingRight = nil then
    Result := false
  else
    Result := true;
end;

function TCTTNode.TaskCount: Integer;
begin
  Result := 1;
  if not (Child = nil) then
    Result := Result + Child.TaskCount;
  if not (SiblingRight = nil) then
    Result := Result + SiblingRight.TaskCount;
end;

end.
