/*
    TaskLib; a Commandline ConcurTaskTree Processing Tool (http://lumumba.luc.ac.be/kris/research/tasklib)
    Copyright (C) 2003  Kris Luyten (kris.luyten@luc.ac.be)
	                     Expertise Centre for Digital Media (http://edm.luc.ac.be)
								Limburgs Universitair Centrum

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of	the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package edm.navigation.concurtasktree;

import java.lang.Integer;

import java.util.Enumeration;
import java.util.Vector;

/**
 * This class constructs a Finite State Machine from a ConcurTaskTree and Enabled Task Sets.
 * @author Tim Clerckx
 */
public class FiniteStateMachineConstructor {
	private Vector $transitions = new Vector();
	private Vector $enabledTaskSets = new Vector();
	private CCTTStep $tree = null;
	private int $start = -1;

	//to fill with suitable tasks for left and right steps
	private Vector $leftTasks         = new Vector();
	private Vector $rightTasks        = new Vector();
	private Vector $leftETS           = new Vector();
	private Vector $rightETS          = new Vector();

	private Vector $recLast           = new Vector();
	//to avoid recLast to be overwritten
	private Vector $recConcurLast     = new Vector();
	private Vector $recFirst          = new Vector();
	//to avoid recFirst to be overwritten when calculating disabling transitions
	private Vector $recFirstDisabling = new Vector();


	public FiniteStateMachineConstructor( CCTTStep tree, Enumeration etsCollection ) throws UnsupportedTransitionException {
		$tree = tree;
		$enabledTaskSets.removeAllElements();
		while( etsCollection.hasMoreElements() ){
			EnabledTaskSet ets = (EnabledTaskSet) etsCollection.nextElement();
			$enabledTaskSets.addElement( ets );
		}
		processTree( $tree );
		removeDuplicates();
	}

	public Enumeration getTransitions(){
		return $transitions.elements();
	}

	public int getStartTaskSet() throws UnsupportedTransitionException {
		return findStartTaskSet( $tree );
	}

	public Enumeration getEnabledTaskSets(){
		return $enabledTaskSets.elements();
	}

	private void removeDuplicates(){
		for( int i = 0; i < $transitions.size(); i++ ){
			FSMTransition t = (FSMTransition) $transitions.elementAt(i);
			for( int j = i+1; j < $transitions.size(); j++ ){
				FSMTransition t2 = (FSMTransition) $transitions.elementAt(j);
				if( t.isEqual( t2 ) )
					$transitions.removeElementAt( j );
			}
		}
	}

	/**
	 * Searching the start task set of a subtree.
	 * @param tree The subtree to examine
	 */
	private int findStartTaskSet( CCTTStep tree ) throws UnsupportedTransitionException {
		//find candidate task sets
		CCTTStep firstTask = getRecursiveFirst( tree );
		//User tasks don't belong to enabled task sets
		if( firstTask.getType().equals( CCTTSAXHandler.TASKTYPE_USER ) ){
			return findStartTaskSet( firstTask.getRightSibling() );
		}else{
			Vector candidates = getTaskSets( firstTask );
			if( candidates.size() == 0 ){
				throw new UnsupportedTransitionException( "Start task does not belong to a task set" );
			}else if( candidates.size() == 1 ){
				return ((Integer)candidates.elementAt(0)).intValue();
			}else{
				$recFirst.removeAllElements();
				getAllRecursiveFirsts( tree );
				Vector tasksets = getTaskSets( $recFirst );
				return ((Integer)tasksets.elementAt(0)).intValue();
            /*
				if( tasksets.size() != 1 ){
					throw new UnsupportedTransitionException( "No. of start task sets != 1" );
				}else{
					return ((Integer)tasksets.elementAt(0)).intValue();
				}*/
			}
		}
	}

	private void getAllRecursiveFirsts( CCTTStep step ){
		if( step.isLeaf() ){
			$recFirst.addElement( step );
		}else{
			if( step.getChild(0).getNext().getType().equals( TemporalRelation.CHOICE ) 
					//deze 2 tests zijn onder voorbehoud ;-)
					|| step.getChild(0).getNext().getType().equals( TemporalRelation.INTERLEAVING ) 
					|| step.getChild(0).getNext().getType().equals( TemporalRelation.SYNCHRONIZATION ) ){
				for( Enumeration e = step.getChildren(); e.hasMoreElements(); ){
					getAllRecursiveFirsts( (CCTTStep)e.nextElement() );
				}
			}else{
				getAllRecursiveFirsts( step.getChild(0) );
			}
		}
	}

	/**
	 * Counts how many tasks in the given set does not have a left sibling
	 * @param taskSet The Task Set to examine
	 */
	private int countMostLeftChilds( int taskSet ){
		Enumeration elts = ((EnabledTaskSet)$enabledTaskSets.elementAt( taskSet )).getEnabledTaskSet();
		int count = 0;
		for( ; elts.hasMoreElements(); ){
			if( ((EnabledTaskSetElement)elts.nextElement()).getStep().getLeftSibling() == null )
				count++;
		}
		return count;
	}

	/**
	 * Recursive function to scan the tree for enabling and disabling tasks. When found one, transitions will be extracted immediately.
	 * @param step The root of the subtree to examine
	 */
	private void processTree( CCTTStep step ) throws UnsupportedTransitionException {
		for( int i = 0; i < step.getNrOfChildren(); i++ ){
			CCTTStep child = step.getChild(i);
			if( child.getNext() != null && 
					( child.getNext().getType().equals( TemporalRelation.SEQUENTIAL_ENABLING ) ||
					  child.getNext().getType().equals( TemporalRelation.SEQUENTIAL_ENABLING_INFO ) ) )
				handleEnablingTasks( child, child.getRightSibling() );
			else if( child.getNext() != null && 
					child.getNext().getType().equals( TemporalRelation.DISABLING ) )
				handleDisablingTasks( child, child.getRightSibling() );
			else if( child.isLeaf() && child.getNext() == null && !( child.getPrevious() != null 
							&& child.getPrevious().getType().equals( TemporalRelation.DISABLING ) ) )
				handleExpiringTransitions( child );
			if( child.isLeaf() && child.isIterative() )
				handleIterativeLeaves( child );
				
			processTree( child );
		}
	}

	/**
	 * Process transitions caused by an enabling temporal operator between two tasks.
	 */
	private void handleEnablingTasks( CCTTStep left, CCTTStep right ) throws UnsupportedTransitionException {
		if( left.isLeaf() && right.isLeaf() ){
			handleEnablingTwoLeaves( left, right );
		}else if( left.isLeaf() ){ //right is not a leaf
			handleEnablingLeftLeaf( left, right );
		}else if( right.isLeaf() ){ //left is not a leaf
			handleEnablingRightLeaf( left, right );
		}else{ //left and right aren't leaves
			handleEnablingNoLeaves( left, right );
		}
	}

	private void handleEnablingConcurrent( CCTTStep left, CCTTStep right ){
		$leftTasks.removeAllElements();
		$rightTasks.removeAllElements();
		$leftETS.removeAllElements();
		$rightETS.removeAllElements();

		//add ETS to examine to the vectors
		addLeafsFromSubtree( right, $rightTasks );
		$rightETS.addAll( getTaskSets( getRecursiveFirst(right) ) );
		$recConcurLast.removeAllElements();
		getRecursiveConcurLast( left );
		addLeafsFromSubtree( left, $leftTasks );
		//$leftTasks.addAll( $recLast );
		for( Enumeration e = $recConcurLast.elements(); e.hasMoreElements(); ){
			$leftETS.addAll( getTaskSets( (CCTTStep) e.nextElement() ) );
		}

		addTransitions( left, right );
	}

	private void addLeafsFromSubtree( CCTTStep root, Vector vec ){
		if( root.isLeaf() ){
			vec.addElement( root );
		}else{
			for( Enumeration e = root.getChildren(); e.hasMoreElements(); ){
				addLeafsFromSubtree( ((CCTTStep)e.nextElement()), vec );
			}
		}
	}

	private void addTransitions( CCTTStep left, CCTTStep right ){
		for( Enumeration leftenum = $leftETS.elements(); leftenum.hasMoreElements(); ){
			int leftETSnr = ((Integer)leftenum.nextElement()).intValue();
			EnabledTaskSet leftets = (EnabledTaskSet) $enabledTaskSets.elementAt( leftETSnr );
			//add transition for each ETS in rightETS which are the same except for elements belonging to $leftTasks or $righttasks
			for( Enumeration rightenum = $rightETS.elements(); rightenum.hasMoreElements(); ){
				int rightETSnr = ((Integer)rightenum.nextElement()).intValue();
				EnabledTaskSet rightets = (EnabledTaskSet) $enabledTaskSets.elementAt( rightETSnr );
				if( etsMatch( leftets, rightets ) ){
					//add transitions from leftets to rightets for all elements of leftets in $leftTasks
					for( Enumeration e = $leftTasks.elements(); e.hasMoreElements(); ){
						CCTTStep task = (CCTTStep) e.nextElement();
						if( leftets.contains( task ) && $recConcurLast.contains(task) ){
							FSMTransition trans = new FSMTransition(
								leftETSnr,
							rightETSnr,
								task,
								FSMTransition.ENABLING );
							$transitions.addElement( trans );
						}
					}
				}
			}
		}
	}

	private boolean etsMatch( EnabledTaskSet leftets, EnabledTaskSet rightets ){
		for( Enumeration rightenum = rightets.getEnabledTaskSet(); rightenum.hasMoreElements(); ){
			EnabledTaskSetElement rightetse = (EnabledTaskSetElement) rightenum.nextElement();
			if( !$rightTasks.contains(rightetse.getStep()) ){
//			if( !$recConcurLast.contains(rightetse.getStep()) ){
				//it has to be in the other
				boolean found = false;
				for( Enumeration leftenum = leftets.getEnabledTaskSet(); leftenum.hasMoreElements(); ){
					EnabledTaskSetElement leftetse = (EnabledTaskSetElement) leftenum.nextElement();
					if( leftetse.getStep() == rightetse.getStep() ){
						found = true;
					}
				}
				if( !found ){
					return false;
				}
			}
		}
		return true;
	}

	/**
	 * Searches the leftmost child of the given subtree
	 * @param step The subtree to examine
	 * @return The leftmost child of the given subtree
	 */
	private CCTTStep getRecursiveFirst( CCTTStep step ){
		if( step.isLeaf() )
			return step;
		else
			return getRecursiveFirst( step.getChild(0) );
	}

	/**
	 * Searches the rightmost children of the given subtree. Modifies $recLast.
	 * @param step The subtree to examin
	 */
	private void getRecursiveLast( CCTTStep step ){
		if( step.isLeaf() ){
			if( step.getPrevious() != null && step.getPrevious().getType().equals( TemporalRelation.DISABLING ) )
				getRecursiveLast( step.getLeftSibling() );
			else
				$recLast.addElement( step );
		}else{
			if( step.getChild(0).getNext().getType().equals( TemporalRelation.CHOICE ) 
					//deze 2 tests zijn onder voorbehoud ;-)
					|| step.getChild(0).getNext().getType().equals( TemporalRelation.INTERLEAVING ) 
					|| step.getChild(0).getNext().getType().equals( TemporalRelation.SYNCHRONIZATION ) ){
				for( Enumeration e = step.getChildren(); e.hasMoreElements(); ){
					getRecursiveLast( (CCTTStep)e.nextElement() );
				}
			}else{
				getRecursiveLast( step.getChild( step.getNrOfChildren()-1 ) );
			}
		}
	}

	private void getRecursiveConcurLast( CCTTStep step ){
		if( step.isLeaf() ){
			if( step.getPrevious() != null && step.getPrevious().getType().equals( TemporalRelation.DISABLING ) )
				getRecursiveConcurLast( step.getLeftSibling() );
			else
				$recConcurLast.addElement( step );
		}else{
			if( step.getChild(0).getNext().getType().equals( TemporalRelation.CHOICE ) 
					//deze 2 tests zijn onder voorbehoud ;-)
					|| step.getChild(0).getNext().getType().equals( TemporalRelation.INTERLEAVING ) 
					|| step.getChild(0).getNext().getType().equals( TemporalRelation.SYNCHRONIZATION ) ){
				for( Enumeration e = step.getChildren(); e.hasMoreElements(); ){
					getRecursiveConcurLast( (CCTTStep)e.nextElement() );
				}
			}else{
				getRecursiveConcurLast( step.getChild( step.getNrOfChildren()-1 ) );
			}
		}
	}

	private void handleEnablingNoLeaves( CCTTStep left, CCTTStep right ) throws UnsupportedTransitionException{
		$recLast.removeAllElements();
		getRecursiveLast( left );

		for( Enumeration e = $recLast.elements(); e.hasMoreElements(); ){
			CCTTStep step = (CCTTStep) e.nextElement();
			handleEnablingLeftLeaf( step, right );
		}
				
	}
	
	private void handleEnablingRightLeaf( CCTTStep left, CCTTStep right ){
		$recLast.removeAllElements();
		getRecursiveLast( left );
		Vector rightTS = getTaskSets( right );
				
		for( Enumeration e = $recLast.elements(); e.hasMoreElements(); ){
			CCTTStep currentStep = (CCTTStep) e.nextElement();
			//check if there isn't an iterative task in the path between currentStep and left
			if( noIterationInPath( left, currentStep ) ){
				Vector leftTS = getTaskSets( currentStep );
	
				if( rightTS.size() == 1 ){
					for( Enumeration en = leftTS.elements(); en.hasMoreElements(); ){
						FSMTransition trans = new FSMTransition(
								( (Integer)en.nextElement() ).intValue(),
								( (Integer)rightTS.elementAt(0) ).intValue(),
								currentStep,
								FSMTransition.ENABLING );
						$transitions.addElement( trans );
					}
				}else{
					handleEnablingConcurrent( left, right );
				}
			}		
		}
	}

	private boolean noIterationInPath( CCTTStep parent, CCTTStep child ){
		if( child.isIterative() ){
			return false;
		}else if( parent == child ){
			return true;
		}else{
			return noIterationInPath( parent, child.getParent() );
		}
	}
	
	private void handleEnablingLeftLeaf( CCTTStep left, CCTTStep right ) throws UnsupportedTransitionException{
		Vector leftTS  = getTaskSets( left );
		Vector rightTS = getTaskSets( getRecursiveFirst( right ) );

		if( leftTS.size() == 1 ){
/*			for( Enumeration e = rightTS.elements(); e.hasMoreElements(); ){
				FSMTransition trans = new FSMTransition(
						( (Integer)leftTS.elementAt(0) ).intValue(),
						( (Integer)e.nextElement()     ).intValue(),
						left,
						FSMTransition.ENABLING );
				$transitions.addElement( trans );
			}*/
			FSMTransition trans = new FSMTransition(
				( (Integer)leftTS.elementAt(0) ).intValue(),
				findStartTaskSet( right ),
				left,
				FSMTransition.ENABLING );
			$transitions.addElement( trans );
		}else{
			handleEnablingConcurrent( left, right );
		}
	}

	private void handleEnablingTwoLeaves( CCTTStep left, CCTTStep right ) throws UnsupportedTransitionException {	
		Vector leftTS  = getTaskSets( left );
		Vector rightTS = getTaskSets( right );
		
		if( leftTS.size() == 0 || rightTS.size() == 0 )
			throw new UnsupportedTransitionException( "Left and/or Right step not found in Enabled Task Sets" );
		else if( leftTS.size() == 1 ){
			for( Enumeration e = rightTS.elements(); e.hasMoreElements(); ){
				FSMTransition trans = new FSMTransition( 
						( (Integer)leftTS.elementAt(0) ).intValue(),
						( (Integer)e.nextElement()     ).intValue(),
						left,
						FSMTransition.ENABLING );
				$transitions.addElement( trans );
			}
		}else{
			handleEnablingConcurrent( left, right );
		}
	}

	/**
	 * Get the task sets which contain task.
	 * @param task The task which has to be found in the task sets.
	 */
	private Vector getTaskSets( CCTTStep task ){
		Vector targetTasks = new Vector();
		for( int i = 0; i < $enabledTaskSets.size(); i++ ){
			EnabledTaskSet ets = (EnabledTaskSet) $enabledTaskSets.elementAt(i);
			if( ets.contains( task ) ){
				targetTasks.add( new Integer( i ) );
			}
		}
		return targetTasks;
	}

	private Vector getTaskSets( Vector tasks ){
		Vector targetTasks = new Vector();
		for( int i = 0; i < $enabledTaskSets.size(); i++ ){
			EnabledTaskSet ets = (EnabledTaskSet) $enabledTaskSets.elementAt(i);
			boolean found = true;
			int j = 0;
			while( found && j < tasks.size() ){
				if( !ets.contains( (CCTTStep)tasks.elementAt(j) ) ){
					found = false;
				}
				j++;
			}
			if( found ){
				targetTasks.add( new Integer( i ) );
			}

		}
		return targetTasks;
	}

	/**
	 * Process transitions caused by a disabling temporal operator between two tasks.
	 */
	private void handleDisablingTasks( CCTTStep left, CCTTStep right ) throws UnsupportedTransitionException {
		CCTTStep itAnc = getIterativeAncestor( left );
		if( itAnc != null ){ 
			//Transition from TS containing right to start of iterative task 
			//new
			//TODO: implement this for 2 other cases!!!
//			System.out.println("WEZENERIN!!!, right: "+right.getIdentifier());
			$recFirstDisabling.removeAllElements();
			getAllRecursiveFirstsDisabling( right );
			for( Enumeration eout = $recFirstDisabling.elements(); eout.hasMoreElements(); ){
				CCTTStep rightone = (CCTTStep) eout.nextElement();
//				System.out.println("Step: "+rightone.getIdentifier());
				
				Vector rightTS = getTaskSets( rightone );
				for( Enumeration e = rightTS.elements(); e.hasMoreElements(); ){
					FSMTransition trans = new FSMTransition( 
							( (Integer)e.nextElement() ).intValue(),
							findStartTaskSet( itAnc ),
							rightone,
							FSMTransition.DISABLING );
					$transitions.addElement( trans );
				}
			}
			//handleEnablingTasks( right, itAnc );
		}else{
			//No iterative ancestor => find next TS or if none: accept state
			CCTTStep enAnc = getEnablingAncestor( left );
			if( enAnc == null ){//accept state
				Vector rightTS = getTaskSets( right );
				for( Enumeration e = rightTS.elements(); e.hasMoreElements(); ){
					FSMTransition trans = new FSMTransition( 
							( (Integer)e.nextElement() ).intValue(),
							-1,
							right,
							FSMTransition.DISABLING );
					$transitions.addElement( trans );
				}
				
			}else{
				Vector rightTS = getTaskSets( right );
				for( Enumeration e = rightTS.elements(); e.hasMoreElements(); ){
					FSMTransition trans = new FSMTransition( 
							( (Integer)e.nextElement() ).intValue(),
							findStartTaskSet( enAnc.getRightSibling() ),
							right,
							FSMTransition.DISABLING );
					$transitions.addElement( trans );
				}
			}
		}
	}

	private void getAllRecursiveFirstsDisabling( CCTTStep step ){
//		System.out.println("getall....");
		if( step.isLeaf() ){
//			System.out.println("leaf found: "+step.getIdentifier());
			$recFirstDisabling.addElement( step );
		}else{
			if( step.getChild(0).getNext().getType().equals( TemporalRelation.CHOICE ) 
					//deze 2 tests zijn onder voorbehoud ;-)
					|| step.getChild(0).getNext().getType().equals( TemporalRelation.INTERLEAVING ) 
					|| step.getChild(0).getNext().getType().equals( TemporalRelation.SYNCHRONIZATION ) ){
//				System.out.println("CHOICE OR CONCURRENT: "+step.getIdentifier());
				for( Enumeration e = step.getChildren(); e.hasMoreElements(); ){
					getAllRecursiveFirstsDisabling( (CCTTStep)e.nextElement() );
				}
			}else{
//				System.out.println("other:"+step.getIdentifier());
				getAllRecursiveFirstsDisabling( step.getChild(0) );
			}
		}
	}


	/**
	 * Returns the first ancestor of a node with the Iteration property. Null will be returned if there isn't one.
	 * @param step The node to examine.
	 * @return The first Iterative Ancestor.
	 */
	private CCTTStep getIterativeAncestor( CCTTStep step ){
		if( step.getParent() == null ){ //root found
			return null;
		}else if( step.getParent().isIterative() ){
			return step.getParent();
		}else{
			return getIterativeAncestor( step.getParent() );
		}
	}

	/**
	 * Returns first ancestor with enabling operator on the right hand side. If none is found, null will be returned.
	 * @param step The node to examine.
	 * @return The first ancestor with enabling operator on the right hand side.
	 */
	private CCTTStep getEnablingAncestor( CCTTStep step ){
		if( step.getParent() == null ){ //root found
			return null;
		}else if( step.getParent().getNext() != null && 
				( step.getParent().getNext().getType().equals( TemporalRelation.SEQUENTIAL_ENABLING ) || 
				step.getParent().getNext().getType().equals( TemporalRelation.SEQUENTIAL_ENABLING_INFO ) ) ){
			return step.getParent();
		}else{
			return getEnablingAncestor( step.getParent() );
		}
	}
	
	private int count = 0;
	
	private void handleExpiringTransitions( CCTTStep step ) throws UnsupportedTransitionException {
		count++;		
		if( getEnablingAncestor( step ) == null ){//else this transition is already taken care of
			CCTTStep itAnc = getIterativeAncestor( step );
			if( itAnc == null ){//accept state
				Vector stepTS = getTaskSets( step );
				for( Enumeration e = stepTS.elements(); e.hasMoreElements(); ){
					int ets = ( (Integer)e.nextElement() ).intValue();
					if( checkExpiring( ets, step ) ){
						FSMTransition trans = new FSMTransition( 
								ets,
								-1,
								step,
								FSMTransition.EXPIRING );
						$transitions.addElement( trans );
					}
				}
			}else{
				Vector stepTS = getTaskSets( step );
				for( Enumeration e = stepTS.elements(); e.hasMoreElements(); ){
					FSMTransition trans = new FSMTransition( 
							( (Integer)e.nextElement() ).intValue(),
							findStartTaskSet( itAnc ),
							step,
							FSMTransition.EXPIRING );
					$transitions.addElement( trans );
				}
			}
		}
	}

	/**
	 * Returns first ancestor with enabling operator on the right hand side. If none is found, null will be returned.
	 * @param step The node to examine.
	 * @return The first ancestor with enabling operator on the right hand side.
	 */
	private CCTTStep getConcurrentAncestor( CCTTStep step ){
		if( step.getParent() == null ){ //root found
			return null;
		}else if( step.getParent().getNext() != null && 
				( step.getParent().getNext().getType().equals( TemporalRelation.INTERLEAVING ) || 
				step.getParent().getNext().getType().equals( TemporalRelation.SYNCHRONIZATION ) ) 
				|| (
					 step.getParent().getPrevious() != null && 
					( step.getParent().getPrevious().getType().equals( TemporalRelation.INTERLEAVING ) || 
					step.getParent().getPrevious().getType().equals( TemporalRelation.SYNCHRONIZATION ) )
				
				   ) ){
			return step;
		}else{
			return getEnablingAncestor( step.getParent() );
		}
	}

	/**
	 * Checks if Expiring Transition may be added
	 */
	private boolean checkExpiring( int etsnum, CCTTStep expstep ){
		CCTTStep concAn = getConcurrentAncestor( expstep );
		if( concAn != null ){ 
			//all non-disabling tasks in de set must be expiring
			EnabledTaskSet ets = (EnabledTaskSet) $enabledTaskSets.elementAt( etsnum );
			Vector tasks = new Vector();
			for( Enumeration e = ets.elements(); e.hasMoreElements(); ){
				EnabledTaskSetElement etse = (EnabledTaskSetElement) e.nextElement();
				if( !(etse.getStep().getPrevious() != null && 
						etse.getStep().getPrevious().getType().equals( TemporalRelation.DISABLING ) ) ){
					tasks.addElement( etse.getStep() );
					if( !isExpiring( etse.getStep() ) ){
						return false;
					}
				}
			}
	   }
		return true;
	}

	private boolean isExpiring( CCTTStep step ){
		if( step.getNext() == null ||
				  step.getNext().getType().equals( TemporalRelation.CHOICE ) ||
				  step.getNext().getType().equals( TemporalRelation.SYNCHRONIZATION ) ||
				  step.getNext().getType().equals( TemporalRelation.INTERLEAVING ) ){
			return true;
		}		
		return false;
	}

	private void handleIterativeLeaves( CCTTStep step ){
		Vector tasksets = getTaskSets( step );
		for( Enumeration e = tasksets.elements(); e.hasMoreElements(); ){
			int ets = ((Integer)e.nextElement()).intValue();
			FSMTransition trans = new FSMTransition( 
					ets,
					ets,
					step,
					FSMTransition.ITERATIVE );
			$transitions.addElement( trans );
		}
	}
}
