/*
    TaskLib; a Commandline ConcurTaskTree Processing Tool (http://lumumba.luc.ac.be/kris/research/tasklib)
    Copyright (C) 2003  Kris Luyten (kris.luyten@luc.ac.be)
	                     Expertise Centre for Digital Media (http://edm.luc.ac.be)
								Limburgs Universitair Centrum

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of	the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/



package edm.navigation.concurtasktree;

import java.io.OutputStream;
import java.io.PrintStream;

import java.util.Vector;
import java.util.Enumeration;

/**
  * Modifies the Original Tree into an equivalent Priority Tree and generates Enabled Task Sets from this tree.
  * 
  * <br /><br /><b>Algorithm:</b>
  * <ol><li>The original tree is transformed into a Priority Tree (this means that the children of each node of 
  * the tree have the same temporal relations between each other.</li>
  * <li>The Enabled Task Sets are being generated:
  * <ul>
  * <li>Start with one Enabled Task Set, containing the root of the tree</li>
  * <li><code>Repeat until</code> each element of the Enabled Task Sets doesn't have any children and doesn't have the first or body prefix</li>
  * <ol>
  * <li>Get an element with the lowest level (so the highest in the tree), that still needs further processing</li>
  * <li>Check the type of Temporal Relation between its children</li>
  * <li>Depending on this type of Temporal Relation, update the Enabled Task Set of the parent, and maybe add some Enabled Task Sets</li>
  * </ol>
  * </ul>
  * </li>
  * </ol>
  * 
  * @author Tim Clerckx
  */
public class TaskSetGenerator{
	private CCTTStep $root;
	private CCTTStep $prioTop;
	private EnabledTaskSetCollection $taskSets = new EnabledTaskSetCollection();
	

	/**
	 * Creates a new TaskSetGenerator.
	 */
	public TaskSetGenerator(){
		this( null );
	}
	
	/**
	 * Creates a new TaskSetGenerator and sets the root of the tree.
	 * @param root The root of the tree
	 */
	public TaskSetGenerator( CCTTStep root ){
		$root = root;
	}

	/**
	 * Set the root of the tree
	 * @param root The root of the tree
	 */
	public void setRoot( CCTTStep root ){
		$root = root;
	}

	/**
	 * Generates Enabled Task Sets. First modifies the tree into an equivalent priority tree.
	 * Then calculates the Enabled Task Sets from the new tree.
	 * @throws edm.navigation.concurtasktree.IllegalTreeException Illegal Tree given to generate Task Sets
	 * @return the Enabled Task Sets as an Enumeration
	 */
	public EnabledTaskSetCollection generateTaskSets() throws IllegalTreeException{
		if( $root != null ){
			$prioTop = (new CCTTPriority()).getPriorityTree( $root );
			return generateEnabledTaskSets( $prioTop );
		}else{
			throw new IllegalTreeException( "root == null" );
		}
	}

	/**
	 * Returns the priority tree of the ConcurTaskTree specification.
	 * @return the priority tree of the ConcurTaskTree specification 
	 **/
	public CCTTStep getPriorityTree(){
		if(($prioTop == null)&&( $root != null ))
			$prioTop = (new CCTTPriority()).getPriorityTree( $root );
		return $prioTop;
	}
	
	/**
	 * Generates the Enabled Task Sets from the constructed Priority Tree.
	 * @return the generated Enabled Task Sets
	 */
	private EnabledTaskSetCollection generateEnabledTaskSets(CCTTStep top){
		$taskSets = new EnabledTaskSetCollection();
		if( $root != null ){
			//generate first taskset -> ROOT
			EnabledTaskSet ets = new EnabledTaskSet();
			ets.addEnabledTaskSetElement( top, EnabledTaskSetElement.NONE );
			$taskSets.add( ets );
			startGeneration();
			deleteUserTasks();
			deleteLonelyDisablingTasks();
			removeDuplicates();
			return $taskSets;
		}else{
			return null;
		}
	}

	/**
	 * Keeps processing the previously generated Enabled Task Sets until they don't contain any childs
	 **/
	private void startGeneration(){
		//start with root
		EnabledTaskSet currentETS = (EnabledTaskSet) $taskSets.elementAt(0);
		EnabledTaskSetElement currentETSE = (EnabledTaskSetElement) currentETS.elementAt(0);
		do
		{
			processETSE( currentETS, currentETSE );
			currentETS = getMaxETSWithLeaves();
			currentETSE = getMaxETSEWithLeaves();
		}while( currentETS != null );
	}

	/**
	 * Returns the ETS with the first ETSE from the highest level from the Enabled Task Sets to be processed
	 * @return the ETS with the first ETSE from the highest level from the Enabled Task Sets to be processed
	 */
	private EnabledTaskSet getMaxETSWithLeaves(){
		EnabledTaskSet currentETS = null;
		EnabledTaskSetElement currentETSE = null;
		for( Enumeration e = $taskSets.elements(); e.hasMoreElements(); ){
			EnabledTaskSet ets = (EnabledTaskSet) e.nextElement();
			for( Enumeration en = ets.getEnabledTaskSet(); en.hasMoreElements(); ){
				EnabledTaskSetElement etse = (EnabledTaskSetElement) en.nextElement();
				if( !etse.getStep().isLeaf() || etse.getType().equals(EnabledTaskSetElement.FIRST) || etse.getType().equals(EnabledTaskSetElement.BODY) ){
					if( currentETS == null || etse.getStep().getTaskDepth() < currentETSE.getStep().getTaskDepth() ){
						currentETS = ets;
						currentETSE = etse;
					}
				}
			}
		}
		return currentETS;
	}	
	
	/**
	 * Returns the first ETSE from the highest level from the Enabled Task Sets to be processed
	 */
	private EnabledTaskSetElement getMaxETSEWithLeaves(){
		EnabledTaskSetElement currentETSE = null;
		for( Enumeration e = $taskSets.elements(); e.hasMoreElements(); ){
			EnabledTaskSet ets = (EnabledTaskSet) e.nextElement();
			for( Enumeration en = ets.getEnabledTaskSet(); en.hasMoreElements(); ){
				EnabledTaskSetElement etse = (EnabledTaskSetElement) en.nextElement();
				if( !etse.getStep().isLeaf() || etse.getType().equals(EnabledTaskSetElement.FIRST) || etse.getType().equals(EnabledTaskSetElement.BODY) ){
					if( currentETSE == null || etse.getStep().getTaskDepth() < currentETSE.getStep().getTaskDepth() ){
						currentETSE = etse;
					}
				}
			}
		}
		return currentETSE;
	}	

	/**
	 * Processes the children of the task (identified by etse) as indicated by the type of temporal relation connecting the children
	 * @param ets the enabled task set containing etse
	 * @param etse the parent task
	 */
	private void processETSE( EnabledTaskSet ets, EnabledTaskSetElement etse ){
		if( !etse.getStep().isLeaf() ){
			String type = etse.getStep().getChild(0).getNext().getType();
			//We need only the type of the temporal relation btween the first and second child. 
			//Because it is a priority tree, the temporal relation between other the siblings will have the same precedence.
				if( type.equals(TemporalRelation.CHOICE) ){
					processChoice( ets, etse );
				}else if( type.equals(TemporalRelation.SYNCHRONIZATION) || type.equals(TemporalRelation.INTERLEAVING) ){
					processParallelDecomposition( ets, etse );
				}else if( type.equals(TemporalRelation.DISABLING) ){
					processDisabling( ets, etse );
				}else if( type.equals(TemporalRelation.SEQUENTIAL_ENABLING) || type.equals(TemporalRelation.SEQUENTIAL_ENABLING_INFO) ){
					processEnabling( ets, etse );
				}else if( type.equals(TemporalRelation.SUSPEND_RESUME) ){
					processSR( ets, etse );
				}
		}else{
			if( etse.getType().equals(EnabledTaskSetElement.FIRST) ){//first(aLeaf) == aLeaf
				etse.setType(EnabledTaskSetElement.NONE);
			}else if( etse.getType().equals(EnabledTaskSetElement.BODY) ){ //body from a leaf -> delete ETS
				$taskSets.removeElement( ets );
			}
		}
	}

	/**
	 * Process an task with children connected with a choice temporal operator.
	 * Choice tasks belong to different enabled task sets, except the "first" subtasks that
	 * have to be in the same enabled task set.
	 * @param ets the enabled task set containing etse
	 * @param etse the parent task (its children are connected with a choice temporal operator)
	 */
	private void processChoice( EnabledTaskSet ets, EnabledTaskSetElement etse ){
		Vector children = getChildrenToProcess( etse, TemporalRelation.CHOICE );
		if( children == null ) //delete ETS (empty body?)
			$taskSets.removeElement( ets );
		else{
			//body belong to different
			EnabledTaskSet newETS;
			for( Enumeration e = children.elements(); e.hasMoreElements(); ){
				newETS = new EnabledTaskSet();
				newETS.addEnabledTaskSetElement( (CCTTStep) e.nextElement(), EnabledTaskSetElement.BODY );
				for( Enumeration en = ets.getEnabledTaskSet(); en.hasMoreElements(); ){
					EnabledTaskSetElement toAdd = (EnabledTaskSetElement) en.nextElement();
					if( toAdd != etse ){
						newETS.addEnabledTaskSetElement( toAdd );
					}
				} 	
				$taskSets.addElement( newETS );
			}

			//first belong to same
			ets.removeElement( etse );
			for( Enumeration e = children.elements(); e.hasMoreElements(); ){
				EnabledTaskSetElement newETSE = new EnabledTaskSetElement( (CCTTStep) e.nextElement(), EnabledTaskSetElement.FIRST );
				ets.addEnabledTaskSetElement( newETSE ); 	
			}
		}
	}

	/**
	 * Process an ETSE for children with a parallel decomposition temporal operator
	 */
	private void processParallelDecomposition( EnabledTaskSet ets, EnabledTaskSetElement etse ){
		Vector children = getChildrenToProcess( etse, TemporalRelation.INTERLEAVING );
		if( children == null ){ //delete ETS (empty body)
			$taskSets.removeElement( ets );
		}else{
			//Replace parent with all of its children
			ets.removeElement( etse );
			for( Enumeration e = children.elements(); e.hasMoreElements(); ){
				CCTTStep step =  (CCTTStep) e.nextElement();
				ets.addEnabledTaskSetElement( step , EnabledTaskSetElement.NONE );
			}
		}
	}


	
	/**
	 * Process an ETSE for children with an enabling temporal operator, assuming that no 2 disabling operators follow each other
	 */
	private void processDisabling( EnabledTaskSet ets, EnabledTaskSetElement etse ){
		Vector children = getChildrenToProcess( etse, TemporalRelation.DISABLING );
		if( children == null ){ //delete ETS (empty body)
			$taskSets.removeElement( ets );
		}else{
			CCTTStep first = (CCTTStep) children.elementAt(0);
			CCTTStep second = (CCTTStep) children.elementAt(1);

			EnabledTaskSet newETS = new EnabledTaskSet();
			second.setNextETS(newETS);
			newETS.addEnabledTaskSetElement( second, EnabledTaskSetElement.BODY );
			for( Enumeration en = ets.getEnabledTaskSet(); en.hasMoreElements(); ){
				EnabledTaskSetElement toAdd = (EnabledTaskSetElement) en.nextElement();
				if( toAdd != etse )
					newETS.addEnabledTaskSetElement( toAdd );
			}
			$taskSets.addElement( newETS );	

			ets.removeElement( etse );
			ets.addEnabledTaskSetElement( first, EnabledTaskSetElement.NONE );
			ets.addEnabledTaskSetElement( second, EnabledTaskSetElement.FIRST );
		}
	}

	/**
	 * Process an ETSE for children with a suspend/resume temporal operator, assuming that no 2 S/R operators follow each other
	 * @param ets  
	 * @param etse
	 */
	private void processSR( EnabledTaskSet ets, EnabledTaskSetElement etse ){
		Vector children = getChildrenToProcess( etse, TemporalRelation.DISABLING );
		if( children == null ){ //delete ETS (empty body)
			$taskSets.removeElement( ets );
		}else{
			CCTTStep first = (CCTTStep) children.elementAt(0);
			CCTTStep second = (CCTTStep) children.elementAt(1);

			EnabledTaskSet newETS = new EnabledTaskSet();
			second.setNextETS(newETS);
			newETS.addEnabledTaskSetElement( second, EnabledTaskSetElement.BODY );
			for( Enumeration en = ets.getEnabledTaskSet(); en.hasMoreElements(); ){
				EnabledTaskSetElement toAdd = (EnabledTaskSetElement) en.nextElement();
				if( toAdd != etse ){
					newETS.addEnabledTaskSetElement( toAdd );
				}
			}
			$taskSets.addElement( newETS );
			
			ets.removeElement( etse );
			ets.addEnabledTaskSetElement( first, EnabledTaskSetElement.NONE );
			ets.addEnabledTaskSetElement( second, EnabledTaskSetElement.FIRST );
		}
	}

	/**
	 * Process an ETSE for children with an enabling temporal operator
	 * @param ets the enabled task set containing etse
	 * @param etse the parent task (its children are connected with a choice temporal operator)
	 */
	private void processEnabling( EnabledTaskSet ets, EnabledTaskSetElement etse ){
		Vector children = getChildrenToProcess( etse, TemporalRelation.SEQUENTIAL_ENABLING );

		//FIRST
		if( etse.getType().equals(EnabledTaskSetElement.FIRST) ){
			CCTTStep child = (CCTTStep) children.elementAt(0); //Isn't this already done in getChildrenToProcess???
			ets.removeElement( etse );
			EnabledTaskSet newETS = new EnabledTaskSet();
			newETS.addEnabledTaskSetElement( child, EnabledTaskSetElement.BODY );
			addAppropriateTasks( newETS, ets, etse );
			ets.addEnabledTaskSetElement( child, EnabledTaskSetElement.FIRST );
			$taskSets.addElement( newETS );
		//BODY
		}else if(( etse.getType().equals(EnabledTaskSetElement.BODY) )){
			//if(checkValidDisablingParent(etse,ets)){
				for( Enumeration e = children.elements(); e.hasMoreElements(); ){
					EnabledTaskSet newETS = new EnabledTaskSet();
					CCTTStep child = (CCTTStep) e.nextElement();
					newETS.addEnabledTaskSetElement( child, EnabledTaskSetElement.NONE );
					addAppropriateTasks( newETS, ets, etse );
					$taskSets.addElement( newETS );
				}
			//}
			$taskSets.removeElement( ets );
		//NONE
		}else{
			//Make ETS's from the old by copying them and replace the parent by each child
			EnabledTaskSet newETS = null, newETS2 = null;
			for( Enumeration e = children.elements(); e.hasMoreElements(); ){
				CCTTStep step  = (CCTTStep) e.nextElement(); 
				if (newETS == null)
					newETS = new EnabledTaskSet();
				if (e.hasMoreElements())
					newETS2 = new EnabledTaskSet();
				else
					newETS2 = etse.getStep().getNextETS();
				newETS.addEnabledTaskSetElement( step , EnabledTaskSetElement.NONE );
				step.setNextETS(newETS2);
				for( Enumeration en = ets.getEnabledTaskSet(); en.hasMoreElements(); ){
					EnabledTaskSetElement toAdd = (EnabledTaskSetElement) en.nextElement();
					if((toAdd != etse)&&(isAppropriateToAdd_Enabling(step, etse.getStep(), toAdd.getStep())))
						newETS.addEnabledTaskSetElement( toAdd );
				}	
				$taskSets.addElement( newETS );
				newETS = newETS2;
			}
			$taskSets.removeElement( ets );
		}
	}
	
	private void addAppropriateTasks( EnabledTaskSet newETS, EnabledTaskSet oldETS, EnabledTaskSetElement parentETSE ){
		for( Enumeration en = oldETS.getEnabledTaskSet(); en.hasMoreElements(); ){
			EnabledTaskSetElement toAdd = (EnabledTaskSetElement) en.nextElement();
			if( toAdd != parentETSE && isAppropriateToAdd( parentETSE.getStep(), toAdd.getStep() ) )
				newETS.addEnabledTaskSetElement( toAdd );			
		}
	}

	private boolean isAppropriateToAdd( CCTTStep parent, CCTTStep toAdd ){
		boolean test = testConcurrent(parent,toAdd);	
		return ( testDisabling(parent,toAdd) || test );
	}

	/**
	 * When processing an enabling operator, this method checks whether the toAdd subtask can be added
	 * to the same ETS as current is located in.
	 * @param current the CCTTStep which the algorithm is currently processing
	 * @param parent the parent CCTTStep of current
	 * @param toAdd the CCTTStep that may be added to an ETS of current
	 * @return true if toAdd and current can be in the same ETS, otherwise false
	 * @see edm.navigation.concurtasktree.CCTTStep
	 * @see #processEnabling
	 **/
	private boolean isAppropriateToAdd_Enabling(CCTTStep current, CCTTStep parent, CCTTStep toAdd ){
		//check wether toAdd is a task located on the left-hand-side and is involved in a choice relation:
		if(toAdd.leftOf(current) && current.ancestorRelation(TemporalRelation.CHOICE, toAdd)){
			//if current is on the right of a enabling, toAdd can not appear in the same ets
			if(current.getPrevious()!=null)
				return false;
		}
		return true;
	}

	/**
	 * When processing the ``body'' part of an enabling operator, this method checks whether the
	 * curent ETS is valid: it is not valid when its ancestors are involved in
	 * a disabling relation and the left hand side of this relation has children which are included
	 * in this set. <b>Please replace me by a more comprehensible procedure.</b>
	 * @param parent the parent having children involved in an Enabling relation
	 * @param set the set to which parent belongs and that is currently being processed
	 * @return true if it is a valid set, false if the set is not valid (the set will be removed)
	 * @see #processEnabling
	 **/
	private boolean checkValidDisablingParent(EnabledTaskSetElement parent, EnabledTaskSet set){
		TemporalRelation tr = parent.getStep().firstAncestorRelation(TemporalRelation.DISABLING);
		if(tr == null)
			return true;
		CCTTStep left  = (CCTTStep)tr.getLeftTask();
		CCTTStep right = (CCTTStep)tr.getRightTask();
		if((parent.getStep().isChild(right))||parent.getStep()==right){
			for(Enumeration e = left.getChildren(); e.hasMoreElements() ; ){				
				if(set.contains((CCTTStep)e.nextElement()))
					return false;
			}
			return true;
		}else{
			return true;
		}
	}


	private boolean testDisabling( CCTTStep parent, CCTTStep toAdd ){
		if ( toAdd.getPrevious() != null && 
			toAdd.getPrevious().getType().equals(TemporalRelation.DISABLING) ){
			//task is disabling -> check if it is disabling a parent
			if( disablingParent( parent, toAdd ) ){
				return true;
			}else{
				return false;
			}
		}
		return false;
	}

	private boolean disablingParent( CCTTStep parent, CCTTStep toAdd ){
		if( parent == $prioTop ){
			return false;
		}else if( parent.getNext() != null &&
				parent.getNext().getType().equals(TemporalRelation.DISABLING) && parent.getRightSibling() == toAdd ){
			//we have a match in de path
			return true;
		}else{
			return disablingParent( parent.getParent(), toAdd );
		}
	}

	private boolean testConcurrent( CCTTStep parent, CCTTStep toAdd ){
		if( parent == $prioTop ){
			return false;
		}else if( isConcurrentTask(parent) ){
			if( checkSiblingsForMatch( parent, toAdd ) ){
				return true;
			}else{
				return testConcurrent( parent.getParent(), toAdd );
			}
		}else{
			return testConcurrent( parent.getParent(), toAdd );
		}		
	}

	/**
	 * Evaluates whether or not the given step has a concurrent relationship with its siblings.
	 */
	private boolean isConcurrentTask( CCTTStep step ){
		return ( ( step.getNext() != null && 
				( step.getNext().getType().equals(TemporalRelation.INTERLEAVING) 
				|| step.getNext().getType().equals(TemporalRelation.SYNCHRONIZATION) ) )
			|| ( step.getPrevious() != null && 
				( step.getPrevious().getType().equals(TemporalRelation.INTERLEAVING) 
				  || step.getPrevious().getType().equals(TemporalRelation.SYNCHRONIZATION) ) ) );
	}

	private boolean checkSiblingsForMatch( CCTTStep parent, CCTTStep toAdd ){
		if( toAdd == $prioTop ){
			return false;
		}else if( isConcurrentTask( toAdd ) && areSiblings( parent, toAdd ) ){
			return true;
		}else{
			return checkSiblingsForMatch( parent, toAdd.getParent() );
		}
	}

	private boolean areSiblings( CCTTStep parent, CCTTStep toAdd ){
		//find left sibling
		for( Enumeration e = toAdd.getParent().getChildren(); e.hasMoreElements(); ){
			CCTTStep totest = (CCTTStep) e.nextElement();
			if( totest == parent ){
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Returns the children to process, given an Enabled Task Set Element and the type of temporal relation between its children. 
	 * Depends on whether the first, body, or all the children have to be processed.
	 * @param etse The Enabled Task Set
	 * @param type The type of temporal relation between the ETS's children
	 * @return The children to process as a Vector of CCTTSteps
	 */
	private Vector getChildrenToProcess( EnabledTaskSetElement etse, String type ){
		Vector children = new Vector();
		CCTTStep parent = etse.getStep();

		//NONE
		if( etse.getType().equals(EnabledTaskSetElement.NONE) ){
			Enumeration e = parent.getChildren();
			while( e.hasMoreElements() ){
				children.addElement( (CCTTStep) e.nextElement() );
			}
			
		//FIRST
		}else if( etse.getType().equals(EnabledTaskSetElement.FIRST) ){
			if( type.equals(TemporalRelation.SEQUENTIAL_ENABLING) ){
				children.addElement( parent.getChild(0) );
			}else{
				Enumeration e = parent.getChildren();
				while( e.hasMoreElements() ){
					children.addElement( (CCTTStep) e.nextElement() );
				}
			}
	
		//BODY
		}else{ 
			if( type.equals(TemporalRelation.SEQUENTIAL_ENABLING) ){
				Enumeration e = parent.getChildren();
				e.nextElement(); //first may not be in the body
				while( e.hasMoreElements() ){
					children.addElement( (CCTTStep) e.nextElement() );
				}
			}/*else if(type.equals(TemporalRelation.CHOICE)){
				Enumeration e = parent.getChildren();
				while( e.hasMoreElements() ){
					children.addElement( (CCTTStep) e.nextElement() );
				}
			}*/else{
				children = null; //ETS has to be deleted because there is no body
			}
		}
		return children;
	}

	/**
	 * Deletes the ETS's that only contain disabling tasks
	 */
	private void deleteLonelyDisablingTasks(){
		for( Enumeration e = $taskSets.elements(); e.hasMoreElements(); ){
			EnabledTaskSet ets = (EnabledTaskSet) e.nextElement();
			if( onlyDisablesLeft( ets ) ){
				$taskSets.removeElement( ets );
			}
		}
	}
	
	/**
	 * Deletes the user tasks and maybe the ETS if there are only disabling tasks left
	 */
	private void deleteUserTasks(){
		for( Enumeration e = $taskSets.elements(); e.hasMoreElements(); ){
			EnabledTaskSet ets = (EnabledTaskSet) e.nextElement();
			for( Enumeration en = ets.getEnabledTaskSet(); en.hasMoreElements(); ){
				EnabledTaskSetElement etse = (EnabledTaskSetElement) en.nextElement();
				if( etse.getStep().getType().equals( CCTTSAXHandler.TASKTYPE_USER ) ){
					ets.removeElement( etse );
				}
			}
		}
	}

	/**
	 * Tests if the ets contains only disabling operators
	 * @param ets The Enabled Task Set
	 */
	private boolean onlyDisablesLeft( EnabledTaskSet ets ){
		for( Enumeration e = ets.elements(); e.hasMoreElements(); ){
			EnabledTaskSetElement etse = (EnabledTaskSetElement) e.nextElement();
			if( etse.getStep().getPrevious() == null || !etse.getStep().getPrevious().getType().equals(TemporalRelation.DISABLING) ){
				return false;
			}
		}
		return true;
	}

	private void removeDuplicates(){
		for( int i = 0; i < $taskSets.size(); i++ ){
			EnabledTaskSet ets = (EnabledTaskSet) $taskSets.elementAt(i);
			for( int j = i+1; j < $taskSets.size(); j++ ){
				EnabledTaskSet ets2 = (EnabledTaskSet) $taskSets.elementAt(j);
				if( ets.isEqual( ets2 ) ){
					$taskSets.removeElementAt( j );
					j--;
				}
			}
		}
	}


	/**
	 * Prints the Enabled Task Sets on an OutputStream
	 * @param os The OutputStream to print on
	 */
	public void print( OutputStream os ){
		PrintStream output = new PrintStream(os);
		output.println("Enabled Task Sets: ("+$taskSets.size()+")");
		output.println("========================");
		for( Enumeration e = $taskSets.elements(); e.hasMoreElements(); ){
			EnabledTaskSet ets = (EnabledTaskSet) e.nextElement();
			ets.print( os ); 
  		   output.println();
		}
	} 
}
