/*
    TaskLib; a Commandline ConcurTaskTree Processing Tool (http://lumumba.luc.ac.be/kris/research/tasklib)
    Copyright (C) 2003  Kris Luyten (kris.luyten@luc.ac.be)
	                     Expertise Centre for Digital Media (http://edm.luc.ac.be)
								Limburgs Universitair Centrum

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of	the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package edm.navigation.concurtasktree;

import edm.parsing.AbstractSAXHandler;

import edm.navigation.NavigationReader;
import edm.navigation.concurtasktree.TaskObject;

import org.xml.sax.SAXException;
import org.xml.sax.Attributes;

import edm.TagMapper;


/**
  Parses an XML description from a ConcurTaskTree
  @author Kris Luyten, Tim Clerckx
 **/
public class CCTTSAXHandler extends AbstractSAXHandler{

	public final static String CATEGORY             = "Category";
	public final static String DESCRIPTION          = "Description";
	public final static String FREQUENCY            = "Frequency";
	public final static String IDENTIFIER           = "Identifier";
	public final static String NAME                 = "Name";
	public final static String TEMP_OP_NAME         = "name";
	public final static String PARENT               = "Parent";
	public final static String PRECONDITION         = "Precondition";
	public final static String SIBLINGLEFT          = "SiblingLeft";
	public final static String SIBLINGRIGHT         = "SiblingRight";
	public final static String SUBTASK              = "SubTask";
	public final static String TASK                 = "Task";
	public final static String TASKMODEL            = "TaskModel";
	public final static String TEMPORALOPERATOR     = "TemporalOperator";
	public final static String TIMEPERFORMANCE      = "TimePerformance";
	public final static String TYPE                 = "Type";
	public final static String INPUTACTION          = "InputAction";
	public final static String OUTPUTACTION         = "OutputAction";

	public final static String ITERATIVE            = "Iterative";
	public final static String OPTIONAL             = "Optional";
	public final static String PART_OF_COOPERATION  = "PartOfCooperation";
	public final static String PLATFORM             = "Platform";

	//Object
	public final static String OBJECT               = "Object";
	public final static String INPUT_ACTION         = "InputAction";
	public final static String OUTPUT_ACTION        = "OutputAction";
	public final static String FROM                 = "From";
	public final static String TO                   = "To";

	//TimePerformance
	public final static String MAX                  = "Max";
	public final static String MIN                  = "Min";
	public final static String AVERAGE              = "Average";

	//TaskObject
	public final static String TASKOBJ_NAME         = "name";
	public final static String TASKOBJ_CLASS        = "class";
	public final static String TASKOBJ_TYPE         = "type";
	public final static String TASKOBJ_ACCESS_MODE  = "access_mode";
	public final static String TASKOBJ_CARDINALITY  = "cardinality";

	public final static String TASKTYPE_ABSTRACT    = "abstraction";
	public final static String TASKTYPE_APPLICATION = "application";
	public final static String TASKTYPE_INTERACTION = "interaction";
	public final static String TASKTYPE_USER        = "user";

	//UIComponent 
	public final static String UI					      = "ui";

	private String $currentState = null;
	private int $currentTaskDepth = 0;
	private TemporalRelation $currentRelation;
	private TaskObject $currentTaskObject;
	private boolean $parsingObject = false;

	private CCTTStep $firstStep;
	private CCTTStep $currentStep;
	private CCTTStep $previousStep;
	private CCTTStep $leftSibling;
	private CCTTStep $rightSibling;
	private CCTTStep $parent;

	private NavigationReader $theParser;



	String cn = this.getClass().getName();

	/**
	 * Creates a SAX parser for ConcurTaskTree XML files
	 * @param theParser A back reference to the {@link edm.navigation.NavigationReader NavigationReader}
	 * so the parser can be changed into specialized parsers for subtrees
	 **/
	public CCTTSAXHandler(NavigationReader theParser){
		$theParser = theParser;
		$previousStep = $leftSibling = $rightSibling = $parent = $currentStep = $firstStep = null;
		setSAXParser(theParser.getParser());
	}

	/**
	 * Starts executing the actual User Interface which is associated with
	 * this ConcurTaskTree
	 **/
	public void start(){
		$firstStep.execute();
	}

	/**
	 * Returns the root of the parsed ConcurTaskTree
	 * @return the root of the parsed ConcurTaskTree
	 **/
	public CCTTStep getRoot(){
		return $firstStep;
	}

	/**
	 * Creates a new Task and inserts it into the tree
	 * @param attr The Attribute list from the Task Tag
	 **/
	private void createTask(Attributes attr) throws CCTTParseException{
		String id=attr.getValue(IDENTIFIER);
		String category=attr.getValue(CATEGORY);
		String frequency=attr.getValue(FREQUENCY);
		try{
			//Create a new step
			CCTTStep step = createStep( category, id, frequency );
			
			//TODO replace by a precondition
			if(step==null){
				throw new CCTTParseException(category);
			}

			//Check if iterative Task
			String it = attr.getValue(ITERATIVE);
			if ( it.equals("true") ){
				step.setIterative( true );
			}else{
				step.setIterative( false );
			}

			//Check if optional Task
			String opt = attr.getValue(OPTIONAL);
			if ( opt.equals("true") ){
				step.setOptional( true );
			}else{
				step.setOptional( false );
			}
			
			//Check if PartOfCooperation
			String poc = attr.getValue(PART_OF_COOPERATION);
			if ( poc.equals("true") ){
				step.setPartOfCooperation( true );
			}else{
				step.setPartOfCooperation( false );
			}
			
			//step.setTaskDepth($currentTaskDepth);
         //Attach the step into the CCTTree
         attachStep( step );
		}catch(NullPointerException npe){
			//no prob
			System.err.println("Unexpected error ("+npe.toString()+")\nContact the dygimes team");
			npe.printStackTrace();
		}

	}


	/**
	 * Creates a new Task of the right type
	 * @param category Category of the Task
	 * @param id Id of the Task
	 * @param frequency of the Task
	 * @return The created Task
	 **/
	private CCTTStep createStep( String category, String id, String frequency ){
		CCTTStep step = null;
		if(category.equals(TASKTYPE_ABSTRACT)){
			step = new AbstractTaskStep(id);
			$currentState=TASKTYPE_ABSTRACT;
		}else if(category.equals(TASKTYPE_INTERACTION)){
			step = new InteractionTaskStep(id);
			step.setFrequency(frequency);
			$currentState=TASKTYPE_INTERACTION;
		}else if(category.equals(TASKTYPE_APPLICATION)){
			step = new ApplicationTaskStep(id);
			$currentState=TASKTYPE_APPLICATION;
		}else if(category.equals(TASKTYPE_USER)){
			step = new UserTaskStep(id);
			$currentState=TASKTYPE_USER;
		}
		return step;
	}


	/**
	 * Attaches a step into the CCTTree
	 * @param step The step that should be attached
	 **/
	private void attachStep( CCTTStep step ){
		if($firstStep == null){//first Task: processed the Root of the CCTT
			$firstStep = step;
			$currentStep = $firstStep;
			if ( $firstStep == null ){
				System.err.println("warning: root == null");
			}
		}else{
			//if this step is created a level deeper, it is a child of the previous step
			if($currentTaskDepth > $currentStep.getTaskDepth()){
				$currentStep.addChild(step);
				step.setParent($currentStep);

				//Saving previous step to store in current
				CCTTStep temp = $currentStep;
				$currentStep = step;
				$currentStep.setPreviousStep( temp );

			}else if($currentTaskDepth == $currentStep.getTaskDepth()){
				//else, it is a sibling of the previous step, and has the same parent as the previous step
				//attach the step to the $currentStep, and move the $currentStep one further (down)

				$currentStep.attachNext($currentRelation);
				step.setParent($currentStep.getParent());
				step.attachPrevious($currentRelation);
				$currentRelation.setRightTask( step );

/*	
 *	set sibling relations <-- This is a redundant association, the TemporalRelation already takes
 *	care of sibling relations!!!
				$currentStep.setRightSibling( step );
				step.setLeftSibling( $currentStep );
*/
				step.setParent( $currentStep.getParent() );
				$currentStep.getParent().addChild( step );

				//Saving previous step to store in current
				CCTTStep temp = $currentStep;
				$currentStep = step;
				$currentStep.setPreviousStep( temp );

			}else{
				//returning to a higher level in the tree
				//so we are processing a sibling on a higher level now

				//calculate how many levels to go up in the tree
				int previousDepth = $currentStep.getTaskDepth();
				int difference = previousDepth-$currentTaskDepth;

				CCTTStep parent = $currentStep.getParent();

				//move up in the tree until we've got the right parent
				for ( int i=0; i<difference; i++ ){
					parent = parent.getParent();
				}

				//append new child
				parent.addChild(step);
				step.setParent(parent);

				//set sibling relations
				
				int children = parent.getNrOfChildren();
				CCTTStep leftSibling = parent.getChild(children-2);
/*				step.setLeftSibling( leftSibling );
				leftSibling.setRightSibling( step );
*/
				//set temporal relations
				TemporalRelation tr = leftSibling.getNext();
				step.attachPrevious( tr );
				tr.setRightTask( step );

				//Saving previous step to store in current
				CCTTStep temp = $currentStep;
				$currentStep = step;
				$currentStep.setPreviousStep( temp );
			}
		}
	}

	/**
	 * Processes an Object required to perform a Task
	 * @param attr The Attribute list from the Object Tag
	 **/
	private void processTaskObject(Attributes attr) throws CCTTParseException{
		try{
			$parsingObject = true;

            //make new TaskObject and set the attributes
            String name = attr.getValue( TASKOBJ_NAME );
            $currentTaskObject = new TaskObject( name );
            $currentTaskObject.setClassName( attr.getValue( TASKOBJ_CLASS ) );
            $currentTaskObject.setType( attr.getValue( TASKOBJ_TYPE ) );
            $currentTaskObject.setAccessMode( attr.getValue( TASKOBJ_ACCESS_MODE ) );
            $currentTaskObject.setCardinality( attr.getValue( TASKOBJ_CARDINALITY ) );
          }catch(NullPointerException npe){
            //no prob
            System.err.println("Unexpected error ("+npe.toString()+")\nContact the edm team");
            npe.printStackTrace();
          }
        }


	/**
	 * Processes a temporal relation
	 * @param trelname type of the temporal relation
	 **/
	private void processTemporalRelation(String trelname)
		throws NoSiblingException, UnsupportedRelationException{
		$currentRelation = RelationsFactory.getInstance().createRelation($currentStep, trelname);
		$currentStep.attachNext($currentRelation);
		//$currentRelation.setTaskDepth($currentTaskDepth);
	}

	public void startDocument() throws SAXException{
	}

	public void endDocument() throws SAXException{
	}





	public void startElement(String nameSpaceURL, String localName, String qName, Attributes attr)
											 throws SAXException{
		resetContents();
		try{
			if(localName.equals(TagMapper.CONSTRAINTS)){
			}else if(localName.equals(TASK)){
				if(attr!=null)
					createTask(attr);
			}else if(localName.equals(SUBTASK)){
				$currentTaskDepth++;
				//make current task the new parent task
				$parent = $currentStep;
			}else if(localName.equals(TEMPORALOPERATOR)){
				try{
					processTemporalRelation(attr.getValue(TEMP_OP_NAME));
				}catch(NoSiblingException nse){
					//will not occur here!
				}catch(UnsupportedRelationException usre){
					System.err.println("Unsupported Temporal relation detected:" + usre.toString());
				}
			}else if( localName.equals(OBJECT) ){
				if(attr!=null)
					processTaskObject( attr );
			}else if( localName.equals(INPUT_ACTION) ){
			}else if( localName.equals(OUTPUT_ACTION) ){
			}else if(localName.equals(UI)){
			}
		}catch(CCTTParseException ccttpe){
			System.err.println(ccttpe.toString());
			System.err.println("Please check whether the ConcurTaskTree XML file is Valid");
			System.exit(0);
		}
	}


	public void endElement(String nameSpaceURL, String localName, String qName) throws SAXException{
		if(localName.equals(TASK)){
		}else if(localName.equals(SUBTASK)){
			$currentTaskDepth--;
			$parent = $parent.getParent();
		}else if(localName.equals(NAME)){
			$currentStep.setName(getContents().toString());
		}else if(localName.equals(TYPE)){
			$currentStep.setType( getContents().toString() );
		}else if(localName.equals(DESCRIPTION)){
			$currentStep.setDescription( getContents().toString() );
		}else if( localName.equals(PLATFORM) ){
			//Platform can be a child tag of either Task or Object
			if ( $parsingObject ){
				$currentTaskObject.addPlatform( getContents().toString() );
			}else{
				$currentStep.addPlatform( getContents().toString() );
			}
		}else if( localName.equals(MAX) ){
			$currentStep.setTimePerfMax( getContents().toString() );
		}else if( localName.equals(MIN) ){
			$currentStep.setTimePerfMin( getContents().toString() );
		}else if( localName.equals(AVERAGE) ){
			$currentStep.setTimePerfAverage( getContents().toString() );
		}else if( localName.equals(OBJECT) ){
			//attach the current Task Object to the current Task
			$parsingObject = false;
			$currentStep.addTaskObject( $currentTaskObject );
		}
	}


	public void characters(char[] ch, int start, int length) throws SAXException{
		writeContents(ch,start,length);
	}	
}
