/*
    TaskLib; a Commandline ConcurTaskTree Processing Tool (http://lumumba.luc.ac.be/kris/research/tasklib)
    Copyright (C) 2003  Kris Luyten (kris.luyten@luc.ac.be)
	                     Expertise Centre for Digital Media (http://edm.luc.ac.be)
								Limburgs Universitair Centrum

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of	the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/


package edm.navigation.concurtasktree;

import edm.navigation.Relation;
import edm.navigation.Step;

/**
  * Presents a temporal relation between two tasks
  * @author Kris Luyten
  */
public class TemporalRelation extends Relation implements Cloneable{

/** Temporal Relation: Sequential Sequential Enabling */
	public final static String SEQUENTIAL_ENABLING      = "SequentialEnabling";
	/** Temporal Relation: Sequential Disabling */
	public final static String DISABLING                = "Disabling";
	/** Temporal Relation: Sequential Choice */
	public final static String CHOICE                   = "Choice";
	/** Temporal Relation: Sequential Interleaving */
	public final static String INTERLEAVING             = "Interleaving";
	/** Temporal Relation: Sequential Synchronization */
	public final static String SYNCHRONIZATION          = "Synchronization";
	/** Temporal Relation: Sequential Suspend/Resume */
	public final static String SUSPEND_RESUME           = "SuspendResume";
	/** Temporal Relation: Sequential Sequential Enabling with Information passing */
	public final static String SEQUENTIAL_ENABLING_INFO = "SequentialEnablingInfo";
	/** Temporal Relation: Order Independence */
	public final static String ORDER_INDEPENDENCE = "OrderIndependence";

	
	protected final static String[] TYPES =
		new String[] {	SEQUENTIAL_ENABLING,
						   DISABLING,
						   CHOICE,
						   INTERLEAVING,
						   SYNCHRONIZATION,
						   SUSPEND_RESUME,
						   SEQUENTIAL_ENABLING_INFO};

	public final static String[][] priorityTable = { {CHOICE }, 
							  { INTERLEAVING, SYNCHRONIZATION, ORDER_INDEPENDENCE }, 
							  { DISABLING }, 
							  { SUSPEND_RESUME },
							  { SEQUENTIAL_ENABLING, SEQUENTIAL_ENABLING_INFO } 
							};


	private String $type;
	private int $taskDepth = -1;

	/**
	 * Creates a new TemporalRelation
	 * @param type the type of temporal relation which is created
	 * @throws UnsupportedRelationException when there is no temporal relation for type
	 **/
	public TemporalRelation(String type)
		throws UnsupportedRelationException{
		super();
		//just check whether it is a legal type
		int i=0;
		$type="";
		while(i<TYPES.length){
			if(type.equals(TYPES[i++]))
				$type=type;
		}
		if((i>=TYPES.length)&&($type==""))
			throw new UnsupportedRelationException(type);

	}

	/**
	 * Creates a new TemporalRelation of the type "type" between stLeft and stRight
	 * @param type the type of temporal relation which is created
	 * @param stLeft the left step involved in the temporal relation
	 * @param stRight the right step involved in the temporal relation
	 * @throws UnsupportedRelationException when there is no temporal relation for type
	 **/
	public TemporalRelation(String type, CCTTStep stLeft, CCTTStep stRight)
		throws UnsupportedRelationException{
			this(type, stLeft);
			setRightTask(stRight);
	}

	/**
	 * Creates a new TemporalRelation of the type "type", with stLeft as a left step
	 * @param type the type of temporal relation which is created
	 * @param stLeft the left step involved in the temporal relation
	 * @throws UnsupportedRelationException when there is no temporal relation for type
	 **/
	public TemporalRelation(String type, CCTTStep stLeft)
		throws UnsupportedRelationException{
			this(type);
			setLeftTask(stLeft);
	}

	public String toString(){
		return getLeftTask() + "-->" + getType() + "-->" + getRightTask();
	}

	/**
	  * The depth in the task decomposition hierarchy where this relation is used
	  * @param tDepth depth in the task decomposition hierarchy
	  */
	public void setTaskDepth(int tDepth){
		$taskDepth = tDepth;
	}

	/**
	  * Returns the depth in the task decomposition hierarchy where this relation is used
	  * @return the depth in the task decomposition hierarchy where this relation is used
	  */
	public int getTaskDepth(){
		return $taskDepth;
	}

	public Step getLeftTask(){
		return getStep(0);
	}

	public Step getRightTask(){
		return getStep(1);
	}

	public String getType(){
		return $type;
	}

	public void setLeftTask(CCTTStep lftStep){
		addStep(lftStep, 0);
	}

	public void setRightTask(CCTTStep rghtStep){
		addStep(rghtStep, 1);
	}

	public Object clone(){
		TemporalRelation tr = null;
		tr = (TemporalRelation)super.clone();
		return tr;
	}


	public boolean priorityIsHigher(TemporalRelation other){
		return getPriority() < other.getPriority();
	}

	public boolean priorityEquals(TemporalRelation other){
		return getPriority() == other.getPriority();
	}

	public boolean priorityIsLower(TemporalRelation other){
		return getPriority() > other.getPriority();
	}

  /**
	 * Calculate the priority of a given type of Temporal Relation
	 * @param type the type of a Temporal Relation
	 * @return the priority of a given type of Temporal Relation (the lower the integer, the higher the priority)
	 */
	private int getPriority(){
		for( int i=0; i<priorityTable.length; i++ ){
			for( int j=0; j<priorityTable[i].length; j++ ){
				if( priorityTable[i][j].equals( getType() ) ){
					return i;
				}
			}
		}
		return priorityTable.length+1;
	}

}
